// Desc: Example code showing how to use D3D lights.
// Copyright (c) 1995-2001 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
unit Unit1;

interface
uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, OmegaImageList, OmegaScreen, OmegaInput, OmegaBitmapFont,
  DXUtil, D3DUtil, OmegaTimer, OmegaUtils, D3DX8, {$I UseD3D8.inc};

type
  // Custom D3D vertex format used by the vertex buffer
  PMyVertex = ^TMyVertex;
  TMyVertex = packed record
    p: TD3DXVector3;       // vertex position
    n: TD3DXVector3;       // vertex normal
  end;
  PMyVertexArray = ^TMyVertexArray;
  TMyVertexArray = array [0..MaxInt div SizeOf(TMyVertex) - 1] of TMyVertex;

const
  D3DFVF_MYVERTEX = (D3DFVF_XYZ or D3DFVF_NORMAL);

type
  TForm1 = class(TForm)
    OmegaScreen1: TOmegaScreen;
    OmegaTimer1: TOmegaTimer;
    OmegaBitmapFont1: TOmegaBitmapFont;
    OmegaInput1: TOmegaInput;
    procedure FormCreate(Sender: TObject);
    procedure OmegaTimer1Timer(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormDestroy(Sender: TObject);
   private
    m_pWallVB: IDirect3DVertexBuffer8; // Vertex buffer for the walls and floor
    m_pSphereMesh: ID3DXMesh;          // Representation of point light
    m_pConeMesh: ID3DXMesh;            // Representation of dir/spot light
    m_light: TD3DLight8;               // Description of the D3D light
    m_n: LongWord;                     // Number of vertices in the ground grid along X
    m_m: LongWord;                      // Number of vertices in the ground grid along Z
    m_nTriangles: LongWord;
    procedure ResetData;
    procedure FrameMove;             // Number of triangles in the ground grid
  end;

var Form1: TForm1;

implementation
{$R *.DFM}

procedure TForm1.FormCreate(Sender: TObject);
begin
 self.ClientWidth := 800;
 self.ClientHeight := 600;
 OmegaScreen1.Width := self.ClientWidth;
 OmegaScreen1.Height := self.ClientHeight;

 OmegaScreen1.Init;

 // Set up wall/floor mesh resolution.  Try changing m_n and m_m to see
 // how the lighting is affected.
 m_n := 32;
 m_m := 32;
 m_nTriangles := (m_n-1)*(m_m-1)*2;
 ResetData;
 DXUtil_Timer(TIMER_START);

 OmegaTimer1.Enabled := True;
end;

procedure TForm1.ResetData;
var
  v: PMyVertexArray;
  dX, dZ{, dU, dV}: Single;
  k, z, x: LongWord;
  mtrl: TD3DMaterial8;
  matIdentity: TD3DXMatrix;
  matView: TD3DXMatrix;
  vFromPt, vLookatPt, vUpVec: TD3DXVector3;
  matProj: TD3DXMatrix;
  fAspect: Single;
  light: TD3DLight8;                 // Description of the D3D light
begin
  // Create a square grid m_n*m_m for rendering the wall
  OmegaScreen1.Device.CreateVertexBuffer(m_nTriangles*3*SizeOf(TMyVertex),
                                            D3DUSAGE_WRITEONLY, D3DFVF_MYVERTEX,
                                            D3DPOOL_MANAGED, m_pWallVB);
  // Fill in the grid vertex data
  m_pWallVB.Lock(0, 0, PByte(v), 0);
  dX := 1.0 / (m_n-1);
  dZ := 1.0 / (m_m-1);
//  dU := 1.0 / (m_n-1); //BAA: never used
//  dV := 1.0 / (m_m-1); //BAA: never used
  k := 0;
  for z:= 0 to (m_m - 2) do
  begin
    for x:= 0 to (m_n - 2) do
    begin
      v[k].p  := D3DXVector3(10 * x*dX, 0.0, 10 * z*dZ);
      v[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p  := D3DXVector3(10 * x*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p  := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p  := D3DXVector3(10 * x*dX, 0.0, 10 * z*dZ);
      v[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p  := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p  := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * z*dZ);
      v[k].n  := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
    end;
  end;
  m_pWallVB.Unlock;

  // Create sphere and cone meshes to represent the lights
  D3DXCreateSphere(OmegaScreen1.Device, 0.25, 20, 20, m_pSphereMesh, nil);
  D3DXCreateCylinder(OmegaScreen1.Device, 0.0, 0.25, 0.5, 20, 20, m_pConeMesh, nil);

  // Set up a material
  D3DUtil_InitMaterial(mtrl, 1.0, 1.0, 1.0);
  with OmegaScreen1.Device do
  begin
   SetMaterial(mtrl);
   // Set miscellaneous render states
   SetRenderState(D3DRS_DITHERENABLE,   iFALSE);
   SetRenderState(D3DRS_SPECULARENABLE, iFALSE);
   // Set the world matrix
   D3DXMatrixIdentity(matIdentity);
   SetTransform(D3DTS_WORLD,  matIdentity);
   // Set the view matrix.
   vFromPt :=   D3DXVector3(-10,  10, -10);
   vLookatPt := D3DXVector3(0.0, 0.0, 0.0);
   vUpVec :=    D3DXVector3(0.0, 1.0, 0.0);
   D3DXMatrixLookAtLH(matView, vFromPt, vLookatPt, vUpVec);
   SetTransform(D3DTS_VIEW, matView);
   // Set the projection matrix
   fAspect := ClientWidth / ClientHeight;
   D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
   SetTransform(D3DTS_PROJECTION, matProj);
   // Turn on lighting.
   SetRenderState(D3DRS_LIGHTING, iTRUE);
   // Enable ambient lighting to a dim, grey light, so objects that
   // are not lit by the other lights are not completely black
   SetRenderState(D3DRS_AMBIENT,
    D3DCOLOR_COLORVALUE(0.25, 0.25, 0.25, 1.0));
   // Set light #0 to be a simple, faint grey directional light so
   // the walls and floor are slightly different shades of grey
   ZeroMemory(@light, SizeOf(light));
   light._Type := D3DLIGHT_DIRECTIONAL;
   light.Direction := D3DXVector3(0.3, -0.5, 0.2);
   with light.Diffuse do
   begin
     r := 0.25;
     g := 0.25;
     b := 0.25;
   end;
   SetLight(0, light);
   // Set light #1 to be a simple, bright directional light to use
   // on the mesh representing light #2
   ZeroMemory(@light, SizeOf(light));
   light._Type := D3DLIGHT_DIRECTIONAL;
   light.Direction := D3DXVector3(0.5, -0.5, 0.5);
   with light.Diffuse do
   begin
    r := 1.0;
    g := 1.0;
    b := 1.0;
   end;
   SetLight(1, light);
   // Light #2 will be the light used to light the floor and walls.  It will
   // be set up in FrameMove() since it changes every frame.
  end;
end;

procedure TForm1.FrameMove;
var
  dwCaps: DWORD;
  x, y, z: Single;
  v: TD3DXVector3;
  m_fTime: single;
begin
  m_fTime := DXUtil_Timer(TIMER_GETAPPTIME);
  ZeroMemory(@m_light, SizeOf(m_light));
  // Rotate through the various light types
  m_light._Type := TD3DLightType(1 + (Round(m_fTime) div 5) mod 3);

  // Make sure the light type is supported by the device.  If
  // D3DVTXPCAPS_POSITIONALLIGHTS is not set, the device does not support
  // point or spot lights, so change light #2's type to a directional light.
  dwCaps := OmegaScreen1.FD3DDEVCAPS8.VertexProcessingCaps;
  if ((dwCaps and D3DVTXPCAPS_POSITIONALLIGHTS) = 0) then
  begin
    if (m_light._Type = D3DLIGHT_POINT) or (m_light._Type = D3DLIGHT_SPOT)
    then m_light._Type:= D3DLIGHT_DIRECTIONAL;
  end;

  // Values for the light position, direction, and color
  x := Sin(m_fTime * 2.000);
  y := Sin(m_fTime * 2.246);
  z := Sin(m_fTime * 2.640);

  m_light.Diffuse.r  := 0.5 + 0.5 * x;
  m_light.Diffuse.g  := 0.5 + 0.5 * y;
  m_light.Diffuse.b  := 0.5 + 0.5 * z;
  m_light.Range      := 100.0;

  case m_light._Type of
    D3DLIGHT_POINT:
    begin
      m_light.Position     := D3DXVec3Scale(v, D3DXVector3(x, y, z), 4.5)^;
      m_light.Attenuation1 := 0.4;
    end;
    D3DLIGHT_DIRECTIONAL:
      m_light.Direction    := D3DXVector3(x, y, z);
    D3DLIGHT_SPOT:
    begin
      m_light.Position     := D3DXVec3Scale(v, D3DXVector3(x, y, z), 2.0)^;
      m_light.Direction    := D3DXVector3(x, y, z);
      m_light.Theta        := 0.5;
      m_light.Phi          := 1.0;
      m_light.Falloff      := 1.0;
      m_light.Attenuation0 := 1.0;
    end;
  end;
  OmegaScreen1.Device.SetLight(2, m_light);
end;

procedure TForm1.OmegaTimer1Timer(Sender: TObject);
var
  matWorld,
  matTrans,
  matRotate: TD3DXMatrix;
  vecFrom, vecAt, vecUp: TD3DXVector3;
  matWorldInv: TD3DXMatrix;
  strLight: String;
begin
 if not OmegaScreen1.CanDraw then exit;
 OmegaInput1.Update;
 if oisbutton1 in OmegaInput1.keyboard.statesclicked then
 begin
  OmegaScreen1.ToggleFullScreen;
  ResetData;
 end;
 OmegaScreen1.ClearScreen(0, 0, 0);
 OmegaScreen1.BeginRender;
 FrameMove;
 with OmegaScreen1.Device do
 begin
  SetTexture(0, nil);
  SetStreamSource(0, m_pWallVB, SizeOf(TMyVertex));
  SetVertexShader(D3DFVF_MYVERTEX);
  // Turn on light #0 and #2, and turn off light #1
  LightEnable(0, TRUE);
  LightEnable(1, FALSE);
  LightEnable(2, TRUE);
  // Draw the floor
  D3DXMatrixTranslation(matTrans, -5.0, -5.0, -5.0);
  D3DXMatrixRotationZ(matRotate, 0.0);
  D3DXMatrixMultiply(matWorld, matRotate, matTrans);
  SetTransform(D3DTS_WORLD, matWorld);
  DrawPrimitive(D3DPT_TRIANGLELIST, 0, m_nTriangles);
  // Draw the back wall
  D3DXMatrixTranslation(matTrans, 5.0, -5.0, -5.0);
  D3DXMatrixRotationZ(matRotate, D3DX_PI/2);
  D3DXMatrixMultiply(matWorld, matRotate, matTrans);
  SetTransform(D3DTS_WORLD, matWorld);
  DrawPrimitive(D3DPT_TRIANGLELIST, 0, m_nTriangles);
  // Draw the side wall
  D3DXMatrixTranslation(matTrans, -5.0, -5.0, 5.0);
  D3DXMatrixRotationX(matRotate,  -D3DX_PI/2);
  D3DXMatrixMultiply(matWorld, matRotate, matTrans);
  SetTransform(D3DTS_WORLD, matWorld);
  DrawPrimitive(D3DPT_TRIANGLELIST, 0, m_nTriangles);
  // Turn on light #1, and turn off light #0 and #2
  LightEnable(0, FALSE);
  LightEnable(1, TRUE);
  LightEnable(2, FALSE);
  // Draw the mesh representing the light
  if (m_light._Type = D3DLIGHT_POINT) then
  begin
   // Just position the point light -- no need to orient it
   D3DXMatrixTranslation(matWorld, m_light.Position.x,
    m_light.Position.y, m_light.Position.z);
  SetTransform(D3DTS_WORLD, matWorld);
  m_pSphereMesh.DrawSubset(0);
 end else
 begin
  // Position the light and point it in the light's direction
  vecFrom := D3DXVector3(m_light.Position.x, m_light.Position.y, m_light.Position.z );
  vecAt := D3DXVector3(m_light.Position.x + m_light.Direction.x,
                          m_light.Position.y + m_light.Direction.y,
                          m_light.Position.z + m_light.Direction.z);
  vecUp := D3DXVector3(0, 1, 0);
  D3DXMatrixLookAtLH(matWorldInv, vecFrom, vecAt, vecUp);
  D3DXMatrixInverse(matWorld, nil, matWorldInv);
  SetTransform(D3DTS_WORLD, matWorld);
  m_pConeMesh.DrawSubset(0);
 end;
 case m_light._Type of
  D3DLIGHT_POINT: strLight:= 'Point Light';
  D3DLIGHT_SPOT:  strLight:= 'Spot Light';
 else
  strLight:= 'Directional Light';
 end;
 end;

 OmegaScreen1.SetBlendMode(bmNormal);
 OmegaBitmapFont1.Print(44, 510, 'Press "space" to toggle fullscreen');
 OmegaBitmapFont1.Print(44, 530, 'Light mode: ' + strLight);

 OmegaScreen1.EndRender;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
 OmegaTimer1.Enabled := False;
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
 if key = 27 then close;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
 m_pWallVB := nil;
 m_pSphereMesh := nil;
 m_pConeMesh := nil;
end;

end.





