unit OmegaTimer;

interface

uses
  Windows, Messages, Classes, Forms;

type
  TOmegaTimer = class(TComponent)
  private
    FWindowHandle: HWND;
    FOldIdleEvent: TIdleEvent;
    FQPCAvailable: Boolean;
    FFrequency: Int64;
    FOldTime, FNewTime: Int64;
    FDelta, FDeltaSecs: Single;
    FFPSOldTime: Int64;
    FFPS, FFPSCurrent: Integer;
    FFPSWarning: Boolean;
    FFPSWarningCount: Integer;
    FFPSWarningFrames: Integer;
    FEnabled: Boolean;
    FTargetFPS, FMinFPS, FMaxFPS: Integer;
    FLossyMode: Boolean;
    FTargetFPSDelta, FMinFPSDelta, FMaxFPSDelta, FLossyFPSDelta: Integer;//Single;
    FOnTimer: TNotifyEvent;
    procedure SetEnabled(const Value: Boolean);
    procedure SetTargetFPS(const Value: Integer);
    procedure AppIdle(Sender: TObject; var Done: Boolean);
    procedure WndProc(var Msg: TMessage);
    procedure UpdateTimer;
    procedure SetMaxFPS(const Value: Integer);
    procedure SetMinFPS(const Value: Integer);
    procedure SetFPSWarningFrames(const Value: Integer);
    procedure SetLossyMode(const Value: Boolean);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Delta: Single read FDelta;
    property DeltaSecs: Single read FDeltaSecs;
    property FPS: Integer read FFPS;
    property FPSWarning: Boolean read FFPSWarning;
  published
    property Enabled: Boolean read FEnabled write SetEnabled;
    property TargetFPS: Integer read FTargetFPS write SetTargetFPS;
    property MinFPS: Integer read FMinFPS write SetMinFPS;
    property MaxFPS: Integer read FMaxFPS write SetMaxFPS;
    property FPSWarningFrames: Integer read FFPSWarningFrames write SetFPSWarningFrames;
    property LossyMode: Boolean read FLossyMode write SetLossyMode;
    property OnTimer: TNotifyEvent read FOnTimer write FOnTimer;
  end;

implementation

uses
  MMSystem;

{ TOmegaTimer }

procedure TOmegaTimer.AppIdle(Sender: TObject; var Done: Boolean);
var
  AbsDelta: Int64;
begin
  if not FEnabled then exit;
  Done := False;

  //Get new time
  if FQPCAvailable then
    QueryPerformanceCounter(FNewTime)
  else
    FNewTime := timeGetTime;

  //Update FPS
  if FNewTime - FFPSOldTime >= FFrequency then
  begin
    FFPS := FFPSCurrent;
    FFPSCurrent := 0;
    FFPSOldTime := FNewTime;
  end;

  //Get time elapsed
  AbsDelta := FNewTime - FOldTime;

  //Shortcut for lossy mode
  if FLossyMode then
  begin
    FDelta := AbsDelta / FTargetFPSDelta;
    FDeltaSecs := AbsDelta / FFrequency;
    if Assigned(FOnTimer) then FOnTimer(Self);
    FOldTime := FNewTime;
    Inc(FFPSCurrent);
  end
  else
  begin
    //Check MinFPS
    if (FMinFPS > 0) and (AbsDelta > FMinFPSDelta) then begin
      AbsDelta := FMinFPSDelta;
      Inc(FFPSWarningCount);
      if FFPSWarningCount >= FFPSWarningFrames then begin
        FFPSWarning := True;
        Dec(FFPSWarningCount);
      end;
    end else begin
      FFPSWarning := False;
      FFPSWarningCount := 0;
    end;

    //Check MaxFPS
    if (FMaxFPS = 0) or (AbsDelta >= FMaxFPSDelta) then begin
      if FTargetFPS > 0 then begin
        if AbsDelta >= FTargetFPSDelta then begin
          FDelta := AbsDelta / FTargetFPSDelta;
          FDeltaSecs := AbsDelta / FFrequency;
          if Assigned(FOnTimer) then FOnTimer(Self);
          FOldTime := FNewTime;
          Inc(FFPSCurrent);
        end;
      end else begin
        FDelta := AbsDelta / FFrequency;
        FDeltaSecs := FDelta;
        if Assigned(FOnTimer) then FOnTimer(Self);
        FOldTime := FNewTime;
        Inc(FFPSCurrent);
      end;
    end;
  end;

  if Assigned(FOldIdleEvent) then FOldIdleEvent(Sender, Done);
end;

procedure TOmegaTimer.WndProc(var Msg: TMessage);
var
  bDummy: Boolean;
begin
  with Msg do
    if Msg = WM_TIMER then
      try
        AppIdle(Self, bDummy);
      except
        Application.HandleException(Self);
      end
    else
      Result := DefWindowProc(FWindowHandle, Msg, wParam, lParam);
end;

procedure TOmegaTimer.UpdateTimer;
begin
  KillTimer(FWindowHandle, 1);
  if FLossyMode then
  begin
    if (FTargetFPS > 0) and FEnabled and Assigned(FOnTimer) then
      SetTimer(FWindowHandle, 1, FLossyFPSDelta, nil);
    Application.OnIdle := nil;
  end
  else
  begin
    if FEnabled and Assigned(FOnTimer) then
      Application.OnIdle := AppIdle
    else
      Application.OnIdle := nil;
  end;
end;

constructor TOmegaTimer.Create(AOwner: TComponent);
begin
  inherited;

  if not (csDesigning in ComponentState) then begin
    //Set up available timing mechanisms on the host machine
    FQPCAvailable := QueryPerformanceFrequency(FFrequency);
    if not FQPCAvailable then begin
      timeBeginPeriod(1);
      FFrequency := 1000;
    end;
  end;

  MinFPS := 10;
  MaxFPS := 120;
  FPSWarningFrames := 10;

  if csDesigning in ComponentState then Exit;

  FWindowHandle := AllocateHWnd(WndProc);

  FOldIdleEvent := Application.OnIdle;
end;

destructor TOmegaTimer.Destroy;
begin
  if not (csDesigning in ComponentState) then
  begin

    if not FQPCAvailable then
      timeEndPeriod(1);

    DeallocateHWnd(FWindowHandle);
  end;
  Application.OnIdle := nil;
  inherited;
end;

procedure TOmegaTimer.SetEnabled(const Value: Boolean);
begin
  if Value = FEnabled then Exit;
  FEnabled := Value;

  if csDesigning in ComponentState then Exit;

  if Value then begin
    if FQPCAvailable then
      QueryPerformanceCounter(FOldTime)
    else
      FOldTime := timeGetTime;
    FFPSOldTime := FOldTime;
    FFPSCurrent := 0;
    FFPS := 0;
  end;

  UpdateTimer;
end;

procedure TOmegaTimer.SetTargetFPS(const Value: Integer);
begin
  if Value = FTargetFPS then Exit;
  FTargetFPS := Value;

  if csDesigning in ComponentState then Exit;

  if FTargetFPS > 0 then begin
    FTargetFPSDelta := FFrequency div FTargetFPS;
    FLossyFPSDelta := 1000 div FTargetFPS;
  end;

  UpdateTimer;
end;

procedure TOmegaTimer.SetMaxFPS(const Value: Integer);
begin
  if Value = FMaxFPS then Exit;
  FMaxFPS := Value;

  if csDesigning in ComponentState then Exit;

  if FMaxFPS > 0 then
    FMaxFPSDelta := FFrequency div FMaxFPS;
end;

procedure TOmegaTimer.SetMinFPS(const Value: Integer);
begin
  if Value = FMinFPS then Exit;
  FMinFPS := Value;

  if csDesigning in ComponentState then Exit;

  if FMinFPS > 0 then
    FMinFPSDelta := FFrequency div FMinFPS;
end;

procedure TOmegaTimer.SetFPSWarningFrames(const Value: Integer);
begin
  if Value = FFPSWarningFrames then Exit;
  FFPSWarningFrames := Value;
  FFPSWarningCount := 0;
  FFPSWarning := False;
end;

procedure TOmegaTimer.SetLossyMode(const Value: Boolean);
begin
  FLossyMode := Value;

  if csDesigning in ComponentState then Exit;

  UpdateTimer;
end;

end.
